import numpy as np
import time
import argparse
from functools import partial
from ur_simple_control.managers import getMinimalArgParser, ControlLoopManager, RobotManager
from ur_simple_control.optimal_control.crocoddyl_optimal_control import get_OCP_args, CrocoIKOCP
from ur_simple_control.basics.basics import jointTrajFollowingPD
from ur_simple_control.util.logging_utils import LogManager
import pinocchio as pin


def get_args():
    parser = getMinimalArgParser()
    parser = get_OCP_args(parser)
    args = parser.parse_args()
    return args


if __name__ == "__main__": 
    args = get_args()
    robot = RobotManager(args)
    Mgoal = robot.defineGoalPointCLI()
    # create and solve the optimal control problem of
    # getting from current to goal end-effector position.
    # reference is position and velocity reference (as a dictionary),
    # while solver is a crocoddyl object containing a lot more information
    reference, solver = CrocoIKOCP(args, robot, Mgoal)
    # we need a way to follow the reference trajectory,
    # both because there can be disturbances,
    # and because it is sampled at a much lower frequency
    jointTrajFollowingPD(args, robot, reference)

    print("final position:")
    print(robot.getT_w_e())

    if not args.pinocchio_only:
        robot.stopRobot()

    if args.visualize_manipulator:
        robot.killManipulatorVisualizer()
    
    if args.save_log:
        robot.log_manager.saveLog()
    #loop_manager.stopHandler(None, None)

